#include <M5StickCPlus.h>
#include <BLEDevice.h>
#include <WiFi.h>
#include <WebServer.h>
#include <ESPmDNS.h>

#define CompanyID 0x0392   // T&D CompanyID

//===== 無線LANアクセスポイント設定(各自の環境を入力) =====
const char* ssid = "xxxxxxxxxxxx";                //(※変更)APのSSID
const char* password = "yyyyyyyyyyyy";     //(※変更)APのパスワード
const char* url_local = "wssnano";          // [url_local].local がブラウザでのアクセス先

typedef struct {
  uint32_t seri;
  char* name;
} Dict;
//===== 名称を表示させたいシリアルNo.と名称を設定(各自の環境を入力) =====
const Dict myDict[8] {
   {0x5F300123,"冷蔵倉庫_南側入口"},    // 入力例{シリアルNo,"表示させたい名称(MAX全角10文字)"}
   {0x5F314567,"冷凍倉庫_北側入口"},    // 入力例{シリアルNo,"表示させたい名称(MAX全角10文字)"}
   {0x5F3289AB,"研究室の冷凍冷蔵庫"},   // 入力例{シリアルNo,"表示させたい名称(MAX全角10文字)"}
   {0x00000000,""},                   // 名称を指定しない場合はシリアルNo.0x00000000 と設定する
   {0x00000000,""},                   // 名称を指定しない場合はシリアルNo.0x00000000 と設定する
   {0x00000000,""},                   // 名称を指定しない場合はシリアルNo.0x00000000 と設定する
   {0x00000000,""},                   // 名称を指定しない場合はシリアルNo.0x00000000 と設定する
   {0x00000000,""},                   // 名称を指定しない場合はシリアルNo.0x00000000 と設定する
};
char buf[16];

struct Device{
  float temp;
  float humi;
  uint32_t seri;
  String model;
  uint8_t tim;
};
struct Device dev[8];

int scanTime = 1;   // In seconds
BLEScan* pBLEScan;
uint8_t adv_cnt;    // アドバタイジングスキャン中に検索した機器数カウント
ulong tm = 0;       // アドバタイジングスキャン間隔タイマ

WebServer server(80);

// Webサーバアクセス時の正常応答をする
void handleRoot(){
  char buf[4000];
  char t_buf[500];
  uint8_t cnt;

  sprintf(buf,
    "<html>\
    <head>\
      <meta http-equiv=\"Refresh\" content=\"60\">\
      <meta charset=\"UTF-8\">\
      <title>WSSnano DataViewer</title>\
      <style media=\"screen\">\
        body{\
          background: #eef4fa;\
          padding:28px 32px;\
          margin:0;\
        }\
        #wrapper{\
          border-radius: 16px;\
          background: #fff;\
          box-shadow: 0px 10px 32px rgba(188, 208, 224, 0.4);\
          padding:30px;\
          margin:0;\
          overflow-x:scroll;\
        }\
        #wrapper h1{\
          font-size: 32px;\
          line-height:1.2;\
          color: #254569;\
          font-weight:700;\
          margin:0;\
          padding:0;\
        }\
        #wrapper table{\
          width:100%%;\
          margin:25px 0 0 0; \
          padding:0;\
          border-collapse: collapse;\
          border-spacing: 0;\
        }\
        #wrapper table th{\
          font-size: 16px;\
          text-align: center;\
          color: #fff;\
          background: #3180c4;\
          padding:5px 23px 0 ;\
          height:40px;\
          box-sizing:border-box;\
          width:20%%;\
          vertical-align:middle;\
          line-height:1.2;\
        }\
        #wrapper table th.th1{\
          width:40%%; \
          text-align:left;\
        }\
        #wrapper table td{\
          font-size: 22px;\
          text-align: center;\
          color: #586b7f;\
          background: #FFF;\
          padding:3px 23px 0 ;\
          height:56px;\
          box-sizing:border-box;\
          vertical-align:middle;\
          line-height:1.2;\
        }\
        #wrapper table td.td1{\
          text-align:left; \
        }\
        #wrapper table tbody tr:nth-child(2n) td{\
          background: #f9fbfd;\
        }\
      </style>\
    </head>\
    <body>\
      <div id='wrapper'>\
        <h1>WSSnano DataViewer</h1>\
      <table>\
        <thead>\
          <tr>\
            <th class='th1'>MODEL</th>\
            <th>NAME</th>\
            <th>Temperature[C]</th>\          
            <th>Humidity[%%]</th>\
          </tr>\
        </thead>\
        <tbody>");

    for(cnt=0 ; cnt < 8; cnt++){
      if(dev[cnt].seri != 0){
        sprintf(t_buf,
          "<tr >\
          <td class='td1'>%s</td>\
          <td>%s</td>\
          ",dev[cnt].model, Ditc_srh(dev[cnt].seri));
          strcat(buf,t_buf);

        if(dev[cnt].temp == 20001){
          sprintf(t_buf,
            "<td>----</td>\
            ");
        }
        else if(dev[cnt].temp == 20000){
          sprintf(t_buf,
            "<td> </td>\
            ");
        }
        else {
          sprintf(t_buf,
            "<td>%3.1f</td>\
            ",dev[cnt].temp);
        }
        strcat(buf,t_buf);

        if(dev[cnt].humi == 20000){
          sprintf(t_buf,
            "<td> </td>\
            </tr>\
            ");
        }
        else if(dev[cnt].humi == 20001){
          sprintf(t_buf,
            "<td>----</td>\
            </tr>\
            ");
        }
        else {
          sprintf(t_buf,
            "<td>%3.0f</td>\
            </tr>\
            ",dev[cnt].humi);
        }
        strcat(buf,t_buf);
      }
    }

    strcat(buf,
      "</tbody>\
      </table>\
    </body>\
    </html>");

  server.send(200,"text/html",buf);
  M5.Lcd.setTextColor(WHITE,BLUE);
  M5.Lcd.drawString(" accessed on \"/\"  ",133,0,2);
}

// Webサーバアクセス時のエラー応答をする
void handleNotFound(){
  server.send(404,"text/plain","File Not Found");
  M5.Lcd.setTextColor(WHITE,RED);
  M5.Lcd.drawString(" File Not Found   ",133,0,2);
}

// シリアルNo → HTTP表示 する名称を決定する
char* Ditc_srh(uint32_t seri){
  uint8_t cnt;
  sprintf(buf,"SN:%08X\0",seri);    // シリアルNo.を文字列に変換する
  Serial.printf("buf:%s\n", buf);
  for(cnt = 0 ; cnt < 8 ; cnt++){
    if(seri == myDict[cnt].seri){   // シリアルNo→名称 の変換データが存在する場合、
      return(myDict[cnt].name);     //   名称をHTML表示
    }
  }
  return(buf);                      // シリアルNo→名称 の変換データがない場合シリアルNoをHTML表示
}

// 受信データの保存領域(8台分)に空きがあるか検索
uint8_t dev_data_srh(uint32_t seri){
  uint8_t cnt, emp_no = 0xff, rtn = 0xff;
  for(cnt = 0 ; cnt < 8 ; cnt++){               // 受信データの保存領域8個分繰り返し
    if(dev[cnt].seri == seri){                  // 受信データの保存領域に既にデータがある場合は
      rtn = cnt;                                // データを上書きする為にそのアドレスを返す。
      break;
    }
    if((dev[cnt].seri == 0)&&(cnt < emp_no)){   // 空領域の最小アドレを探す
      emp_no = cnt;                             // 空領域のアドレスを保存する
    }
  }
  if(rtn == 0xff)rtn = emp_no;                  // 既存データがない場合は空領域アドレスを返す
  return rtn;
}

// シリアルNo.から機種名と機器ごとの情報を取得する
String Find_model(uint32_t seri,uint8_t *flg){
// *flg に機器の下記情報を返す
// bit0: 温度測定 0=無、1=有
// bit1: 湿度測定 0=無、1=有
// bit2: アドバタイズパケットの測定値の位置 0=10、1=8
// bit3-7:未使用
  String rtn;
  uint8_t seri2 = (uint8_t)(seri >> 16);
  rtn = "xxxxx";
  if((seri2 == 0x40)||(seri2 == 0x41)){rtn = "TR41A"; *flg = 0x01;}       // シリアルNo→機種名
  else if((seri2 == 0x42)||(seri2 == 0x43)){rtn = "TR42A"; *flg = 0x01;}
  else if((seri2 == 0x44)||(seri2 == 0x45)){rtn = "TR43A"; *flg = 0x03;}
  else if((seri2 == 0x46)||(seri2 == 0x47)){rtn = "TR32B"; *flg = 0x03;}
  else if((seri2 == 0x2C)||(seri2 == 0x2D)){rtn = "TR41 "; *flg = 0x05;}
  else if((seri2 == 0x2E)||(seri2 == 0x2F)){rtn = "TR42 "; *flg = 0x05;}
  else if((seri2 == 0x30)||(seri2 == 0x31)){rtn = "TR45 "; *flg = 0x05;}
  return(rtn);                      // シリアルNo→model名
}

// アドバタイジングスキャン中の1機器ごとの割り込み
class MyAdvertisedDeviceCallbacks: public BLEAdvertisedDeviceCallbacks {
    void onResult(BLEAdvertisedDevice advertisedDevice) {
      if(adv_cnt++ > 100){
        pBLEScan->stop();
      }
      //Serial.printf("Advertised Device: %s \n", advertisedDevice.toString().c_str());
    }
};

void setup() {
  M5.begin();
  Serial.begin(115200);           // シリアル通信初期設定

  M5.Lcd.setRotation(3);          // LCD表示初期設定
  M5.Lcd.setCursor(0, 0, 4);      // 表示位置とフォントを指定
  M5.Lcd.print("Connecting..");   // 最初にConnecting..と表示する

  WiFi.begin(ssid, password);     // 無線LANアクセスポイント接続開始
  while (WiFi.status() != WL_CONNECTED) { // 接続完了を確認
    M5.Lcd.print(".");                    // 接続途中の表示
    delay(300);                           // 300msec間隔で接続完了を確認
  }
  M5.Lcd.println("\nConnected to");     // DHCPで得られたIPアドレスを表示
  M5.Lcd.println(WiFi.localIP());       // DHCPで得られたIPアドレスを表示

  if(MDNS.begin(url_local)){            // Webサーバアドレス [nanowss.local]
  //  M5.Lcd.println("MDNS started");   // DHCPで得られたIPアドレスを表示
  }
  server.on("/",handleRoot);
  server.onNotFound(handleNotFound);  
  server.begin();
  M5.Lcd.println("WebServer started");  // DHCPで得られたIPアドレスを表示

  BLEDevice::init("");              // Bluetooth Low Energy初期設定
  pBLEScan = BLEDevice::getScan();  // BLEスキャンオブジェクト取得
  pBLEScan->setActiveScan(false);   // パッシブスキャン
  pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks());
  pBLEScan->setInterval(100);
  pBLEScan->setWindow(99);
  M5.Lcd.println("Scanning...");    // アドバタイジングスキャン中の表示

  uint8_t cnt, flg;
  for(cnt = 0 ; cnt < 8 ; cnt++){
    if(myDict[cnt].seri != 0x00000000){
      dev[cnt].model = Find_model(myDict[cnt].seri, &flg);
      dev[cnt].temp = 20001;                        // 温度「----」
      if((flg & 0x02) == 0)dev[cnt].humi = 20000;   // 湿度「    」
      else dev[cnt].humi = 20001;                   // 湿度「----」
      dev[cnt].seri = myDict[cnt].seri;
      dev[cnt].tim = 7;
    }
  }
}

void loop() {
  char seri_tmp[10];
  uint8_t cnt, i, count, flg, dt_adr;
  uint32_t now_seri;

  if(tm == 0)tm = millis() + 30000;     // 更新する時間を30秒後に設定する

  for(cnt = 0 ; cnt < 10 ; cnt++){      // アドバタイジングスキャンは30秒ごとに[1秒x10回]行う
    adv_cnt = 0;                        // アドバタイジングスキャンで検索した機器台数のカウンタ
    BLEScanResults foundDevices = pBLEScan->start(scanTime,false);   // アドバタイジングスキャン開始
    count = foundDevices.getCount();    // count=スキャンで見つけたデバイス数
    for(i = 0 ; i < count ; i++){       // デバイスの個数の回数繰り返す
      BLEAdvertisedDevice d = foundDevices.getDevice(i);
      if(d.haveManufacturerData()){                  
        //### アドバタイジングパケットにManufacturerDataが含まれる場合以下実行 ###
        std::string data = d.getManufacturerData(); // data=受信したManufacturerData
        int manu = data[1] << 8 | data[0];          // manu=受信したCompany ID
        if(manu == CompanyID){  
          //### ManufacturerDataがT&D製品ならば以下実行 ###
          now_seri = ((uint32_t)data[5]<<24)+
            ((uint32_t)data[4]<<16)+((uint32_t)data[3]<<8)+data[2]; // now_siri=受信機器のシリアルNo.
          String mdl_tp = Find_model(now_seri, &flg);
          if(mdl_tp != "xxxxx"){
            uint8_t d_id = dev_data_srh(now_seri);      // 受信データの保存領域の空きを探す
            if(d_id != 0xff){   // 表示用のメモリに空きがあれば保存する
              if((flg & 0x04) == 0)dt_adr = 10;     // アドバタイジングパケットの測定値位置を機種により指定
              else dt_adr = 8;                      // アドバタイジングパケットの測定値位置を機種により指定
              dev[d_id].seri = now_seri;                // ManufacturerDataのシリアルNo.保存
              dev[d_id].temp = data[dt_adr+1] << 8 | data[dt_adr];    //- ManufacturerData →温度データ取得
              if(dev[d_id].temp > 20000)dev[d_id].temp = 20001;
              else dev[d_id].temp = (dev[d_id].temp - 1000) / 10;         // 温度データをXX.X[℃]の値に変換
              dev[d_id].humi = data[dt_adr+3] << 8 | data[dt_adr+2];  //- ManufacturerData →湿度データ取得

              if((flg & 0x02) == 0)dev[d_id].humi = 20000;  // 湿度がない機器の湿度値は20000とする
              else{
                if(dev[d_id].humi > 20000)dev[d_id].humi = 20001;
                else dev[d_id].humi = (dev[d_id].humi - 1000) / 10;       // 湿度データをXX.X[%]の値に変換
              }
              dev[d_id].model = mdl_tp;
              dev[d_id].tim = 7;      // 未受信になった場合の表示を継続する回数を指定
            }
          }
        }
      }
    }
    server.handleClient();
  }

  M5.Lcd.setTextColor(BLACK,CYAN);
  M5.Lcd.drawString("  Receivable device                   ",0,0,2);  // 受信機器リストのタイトル表示
  M5.Lcd.fillRect(0,15,240,160,BLACK);
  uint8_t y_line = 0;
  for(i = 0 ; i < 8 ; i++){            // デバイスの個数の回数繰り返す
    Serial.printf("%s S/N:%8X  ",dev[i].model, dev[i].seri);  //(8)
    Serial.printf("temp:%3.1f humi:%3.0f tim:%d\n",dev[i].temp,dev[i].humi,dev[i].tim); //
    M5.Lcd.setTextColor(WHITE,BLACK);
    if(dev[i].seri != 0){       // 受信データの保存領域にデータが入っている場合表示する
      y_line++;
      M5.Lcd.setCursor(10,15*(y_line),2);               // シリアルNo表示位置とフォントを指定
      M5.Lcd.printf("%s %8X",dev[i].model, dev[i].seri); // シリアルNo表示
      M5.Lcd.setCursor(145,15*(y_line),2);              // 温度の表示位置とフォントを指定
      if(dev[i].temp == 20001)M5.Lcd.printf(" ----");              // 温度を表示
      else M5.Lcd.printf("%3.1fC",dev[i].temp);              // 温度を表示

      if(dev[i].humi != 20000){
        M5.Lcd.setCursor(190,15*(y_line),2);            // 湿度の表示位置とフォントを指定
        if(dev[i].humi == 20001)M5.Lcd.printf(" ----");
        else M5.Lcd.printf("%3.0f%%",dev[i].humi);           // 湿度を表示
      }
    }
    if(dev[i].tim > 0){   // 180sec(30sec x 6)未受信だったら表示リストから消去
      if(--dev[i].tim == 0){
        dev[i].seri = 0; dev[i].temp = 0; dev[i].humi = 0;  // 受信データの保存領域をクリア
        dev[i].model = "";                                   // 受信データの保存領域をクリア
      }
    }
  }
  while(millis() < tm){   // アドバタイジングスキャンの間隔待つ
    server.handleClient();
    delay(100);
  }
  tm = 0;
}